#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>

#include "vector.h"
#include "pixel.h"
#include "list.h"
#include "material.h"

// create a new material description
void material_init(FILE *in,list_t *mats,int attrmax)
{
	material_t	*mat;

  // malloc() a material_t structure, use memset() to
  // initialize it to 0 and store the MAT_COOKIE
  mat = (material_t *)malloc(sizeof(material_t));
  memset((void *)mat,0,sizeof(material_t));
  mat->cookie = MAT_COOKIE;

  // load attributes; unlike the camera, the number of attributes is variable
  material_load_attributes(in,mat);

  // ask list_add to add the material entity to the end of the mats list
  list_add(mats,(void *)mat);
}

// read in material attributes
void material_load_attributes(FILE *in, material_t *mat)
{
	char c,attrname[NAME_LEN];
	int count;

  // assume file pointer has already consumed 'material' token
  assert((count = fscanf(in,"%s",mat->name))==1);

  // consume all chars until we get to '{'
  while((c=fgetc(in)) != EOF && c != '{');

  // loop until we hit '}'
  while((c=fgetc(in)) != EOF && c != '}') {
    // put character back into input stream
    fputc(c,in);

    assert((count = fscanf(in,"%s",attrname))==1);

    if(!strcmp(attrname,"ambient")) pix_read(in,mat->ambient);
    if(!strcmp(attrname,"diffuse")) pix_read(in,mat->diffuse);
    if(!strcmp(attrname,"specular")) pix_read(in,mat->specular);

    // consume whitespace
    while((c=fgetc(in)) != EOF && c != '\n');
  }
}

// return material with requested material name (e.g. yellow)
material_t *material_getbyname(list_t *mats,char *name)
{
	material_t	*mat;

  // reset internal list iterator
  list_reset(mats);

  // basic linear search through list
  while(list_not_end(mats)) {
    mat = (material_t *)list_get_data(mats);
    assert(mat->cookie == MAT_COOKIE);
    if(!strcmp(name,mat->name)) return mat;
    list_next_link(mats);
  }
  return(NULL);
}

// print material object
void material_print(material_t *mat, FILE *out)
{
  assert(mat->cookie == MAT_COOKIE);

  // print out 'material' token and material name
  fprintf(out,"material %s\n",mat->name);
  fprintf(out,"{\n");

  // if we have ambient component, print it out
  if(pix_nonzero(mat->ambient)) {
    fprintf(out,"   ");
    pix_print(out,"ambient",mat->ambient);
  }

  // if we have diffuse component, print it out
  if(pix_nonzero(mat->diffuse)) {
    fprintf(out,"   ");
    pix_print(out,"diffuse",mat->diffuse);
  }

  // if we have specular component, print it out
  if(pix_nonzero(mat->specular)) {
    fprintf(out,"  ");
    pix_print(out,"specular",mat->specular);
  }

  // space out materials by placing extra '\n' at end
  fprintf(out,"}\n\n");
}

// produce a formatted dump of the material list
void material_list_print(list_t *mats, FILE *out)
{
	material_t	*mat;

  // reset internal list iterator
  list_reset(mats);

  // print each material; basic list iteration
  while(list_not_end(mats)) {
    mat = (material_t *)list_get_data(mats);
    material_print(mat,out);
    list_next_link(mats);
  }
}

// name accessor
char* material_getname(material_t *mat)
{
  return mat->name;
}

// ambient accessor
void material_getamb(material_t *mat,drgb_t dst)
{
  assert(mat->cookie == MAT_COOKIE);
  pix_copy(mat->ambient,dst);
}

// diffuse accessor
void material_getdiff(material_t *mat,drgb_t dst)
{
  assert(mat->cookie == MAT_COOKIE);
  pix_copy(mat->diffuse,dst);
}

// specular accessor
void material_getspec(material_t *mat,drgb_t dst)
{
  assert(mat->cookie == MAT_COOKIE);
  pix_copy(mat->diffuse,dst);
}
