#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>

#include "vector.h"
#include "matrix.h"
#include "pixel.h"
#include "list.h"
#include "material.h"
#include "object.h"
#include "plane.h"

// create a new plane object and initialize it
void plane_init(FILE *in,list_t *objs,list_t *mats,int attrmax)
{
	plane_t		*pln;
	object_t	*obj;
	char		c,attrname[NAME_LEN];
	int		count;

  // allocate and initialize plane object
  assert((pln = (plane_t *)malloc(sizeof(plane_t))) != NULL);
  memset(pln,0,sizeof(plane_t));

  // init object (generic structure), simply set obj point to point at pln
  object_init(obj=(object_t *)pln,in,objs,mats);

  // both lines work because plane_t looks like object_t if object_t is
  // declared at top of plane_t structure
  assert((pln = (plane_t *)list_get_data(objs)) != NULL);
  assert(obj->cookie == OBJ_COOKIE);

  // set object type (why isn't this coming from parser?)
  strcpy(obj->type,"plane");

  // overload object's "virtual" functions
  obj->printer = plane_print;
  obj->hits = plane_hits;

  // assume we've consumed all generic object info from file,
  // continue parsing info specific to plane

  // loop until we hit '}'
  while((c=fgetc(in)) != EOF && c != '}') {
    // put character back into input stream
    fputc(c,in);

    assert((count = fscanf(in,"%s",attrname))==1);

    if(!strcmp(attrname,"normal")) vec_read(in,pln->abnormal);
    if(!strcmp(attrname,"point")) vec_read(in,pln->point);

    // consume whitespace
    while((c=fgetc(in)) != EOF && c != '\n');
  }

  // not sure what this does...why does object need to keep track
  // of last hit normal?  and why are we setting it to the last
  // read in normal?
  vec_unit(pln->abnormal,obj->last_normal);
  vec_unit(pln->abnormal,pln->normal);

  // set plane equation (A,B,C,D) with N = (A,B,C) and D = N.point
  pln->ndotq = vec_dot(pln->point,pln->normal);
}

void plane_print(object_t *obj, FILE *out)
{
	// this old statement:
//	plane_t *pln = (plane_t *)obj->priv;
	// is not needed if obj only "masquerades" as object_t *
	// but has more stuff "underneath"
	plane_t *pln = (plane_t *)obj;

  object_print(obj, out);

  fprintf(out,"  ");
  vec_print(out," normal",pln->abnormal);
  fprintf(out,"  ");
  pix_print(out," point ",pln->point);

  // space out objects by placing extra '\n' at end
  fprintf(out,"}\n\n");
}

// determine if a vector projected from location *base in direction *dir
// hits the plane...if so return distance from the base to contact point
double plane_hits(object_t *obj,vec_t base,vec_t dir)
{
	// obj argument is "masquerading" as objecgt_t * but we know better
	plane_t	*pln = (plane_t *)obj;
	double	ndotd,t,ndotb,precision=0.000001;

  assert(obj->cookie == OBJ_COOKIE);

  // get angle between view direction and plane normal
  ndotd = vec_dot(dir,pln->normal);

  // if ndotd == 0, ray is parallel to the plane, no intersection
  if(fabs(ndotd) < precision) return(-1);

  // find distance of ray base to plane: N . b + pln->ndotq
  ndotb = vec_dot(base,pln->normal);

  // here we need to temporarily negate N to find distance along ray to plane
//t = (-1.0 * ndotb + pln->ndotq) / ndotd;
  t = (pln->ndotq - ndotb) / ndotd;

  // if t < 0 then intersection "behind" where ray started from
  if(t <= 0) return(-1);

  // hit point is obtained by moving along dir by t amount
  // and adding to ray base
  vec_scale(t, dir, obj->last_hit);
  vec_sum(obj->last_hit,base,obj->last_hit);

  // if z > 0 then hit point is in front of image plane; invalid intersection
  if(obj->last_hit[2] > 0.0) return(-1);

  return(t);
}
